/*
 
 This file is part of LFT.
 
 The LFT software provided in this Distribution is
 Copyright 2006 VOSTROM Holdings, Inc.
 
 The full text of our legal notices is contained in the file called
 COPYING, included with this Distribution.
 
 Authors:
 
 -  Victor Oppleman <lft@oppleman.com>
 -  Eugene Antsilevitch <ugen@xonix.com>
 
 Other copyrights and former authors:
 
 -  Portions copyright (c) Genuity, Inc. 
 -  Portions copyright (c) Nils McCarthy
 
 */


#include "config/acconfig.h"
#include <assert.h>

#ifdef __CYGWIN__
#define __USE_W32_SOCKETS
#include "windows.h"
#include <ws2tcpip.h>
#include <sys/types.h>
#define LITTLE_ENDIAN 1
#define BYTE_ORDER 1
typedef signed long n_long;
typedef signed short n_short;
typedef signed long n_time;
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <getopt.h>
#include "include/net/if_arp.h"
#include "include/netinet/if_ether.h"
#include "include/netinet/ip.h"
#include "include/netinet/ip_icmp.h"
#include "include/netinet/tcp.h"
#include "include/netinet/udp.h"
#else
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <stdio.h>
#include <math.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/if_ether.h>
#include <assert.h>
#include <netdb.h>
#include <pcap.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#ifdef BSD
#include <machine/limits.h>
#endif
#ifdef BSD_IP_STACK
#include <errno.h>
#include <sys/ioctl.h>
#ifndef DARWIN
#include <net/bpf.h>
#include <pcap-int.h>
#endif
#endif

#ifdef sun
#include <limits.h>
#include <strings.h>
#endif

#endif

#include "lft_ifname.h"
#include "lft_lsrr.h"
#include "whois.h"
#include "lft_queue.h"

const char *version = "2.6b5";  	/* set version string */
const char *version_date = "(04/2006)";	/* date of this version */
const char *appname = "LFT";
static char *hostname = NULL;
static char *hostname_lsrr[9];
static int hostname_lsrr_size = 0;
static int dflag = 0;
static int hop_info_size = 0;
static int hop_info_length = 0;

static int seq_start = 0;       /* generate ISN internally by default */
static int dport = 80;          /* set default destination to tcp/80 HTTP */
static int sport = 53;          /* set default source to tcp/53 dns-xfer */
static int auto_ports = 1;      /* enable port autoselection by default */
static int random_source = 0;   /* disable random source port by default */
static int set_tos = 0;         /* disable set ToS bit by default */

static int maxpacklen = 16 * 1024;  /* maximum user-supplied probe length */
                                    /* Ethernet=1500, GigEther(jumbo)=9000 */
static int minpacklen = 60;     /* minimum user-supplied probe length */
static int userlen = 0;         /* user-requested packet length */
static int payloadlen = 0;      /* the final probe payloadlength */
static int win_len = 32768;

static int timeout_ms = 800;	/* timeout between retries */
static int retry_max = 2;       /* number of retries before giving up */
static int retry_min = 1;       /* minimum number of checks per hop */
static double scatter_ms = 20;	/* milleseconds between sends */
static int ahead_limit = 5;     /* number of probes we can send
                                 * without replies if we don't know
                                 * the number of hops */
static int start_dport = 33434; /* starting port for UDP tracing (traceroute 
                                   compatibility for some targets that care */
static int ttl_limit = 30;      /* max # hops to traverse (highest TTL) */
static int ttl_min = 0;         /* user may request to start at a higher TTL */
static int break_on_icmp = 1;	/* break on icmp other than time exceeded */
static int noisy = 0;           /* disable verbose debug by default */
static int nostatus = 0;        /* print status bar by default */
static int userdevsel = 0;      /* by default, we'll select the device */
static int resolve_names = 1;   /* dns resolution enabled by default */
static int hostnames_only = 0;	/* disable printing of IP addresses */
static int timetrace = 0;       /* disable tracer timing by default */
static int max_net_dev_input = 64; /* only take this much input */
static int adaptive = 0;		/* disable state engine by default */
static int use_udp = 0;         /* use UDP instead of TCP */
static int do_netlookup = 0;    /* disable netname lookup by default */
static int do_aslookup = 0;     /* disable asn lookup by default */
static int use_radb = 0;        /* use RADB instead of pwhois */
static int use_cymru = 0;       /* use Cymru instead of pwhois */
static int use_ris = 0;         /* use RIPE NCC RIS instead of pwhois */

static struct in_addr local_address;
static struct in_addr remote_address;

static struct timeval ts_last_sent, now, begin_time, trace_done_time;
static char time_format[] = "%d-%b-%y %H:%M:%S %Z"; /* time display format */
static char tbuf[128];

static char def_payload[] = "\0\0\0\0\0\0\0\0\0\0"; /* default payload for UDP packets */
static char *payload;

static int send_sock;
static int skip_header_len;

#ifdef __CYGWIN__
static int recv_sock;
static int sock = -1;
#endif

static unsigned short ip_id;
static unsigned char tcp_flags = TH_SYN;

static int num_hops = 0;
static int num_sent = 0;
static int num_rcvd = 0;
static int target_open = 0;
static int target_anomaly = 0;

/* holds the pseudo-header for generating checksums */
struct sumh
{
    unsigned int src;
    unsigned int dst;
    unsigned char fill;
    unsigned char protocol;
    unsigned short len;
};

static char *icmp_messages[] = {
    "endpoint",
    "net unreachable",
    "host unreachable",
    "protocol unreachable",
    "port unreachable",
    "need fragment",
    "source fail",
    "net unknown",
    "host unknown",
    "src isolated",
    "net prohib",
    "host prohib",
    "bad tos/net",
    "bad tos/hst",
    "prohibited",
    "precedence violation",
    "precedence cutoff"
};

/* The actual packet data */
struct trace_packet_s
{
    struct ip ip_hdr;
    struct ip_lsrr lsrr;		/* must go here for ip checksum to work */
    struct tcphdr tcp_hdr;
    struct udphdr udp_hdr;
    int size;
    char *payload;
    int payload_len;
}
trace_packet;

/* Packet container with additional info */
struct trace_packet_info_s
{
    int icmp_type;		/* ICMP_UNREACH code, -1 if RST reply */
    int is_done;		/* is this a final hop? */
    short hopno;
    unsigned int   seq;
    struct timeval sent;
    struct timeval recv;		/* 0 if unreceived */
    struct in_addr hopaddr;     /* IP address of router */
    struct trace_packet_s	packet;
    SLIST_ENTRY(trace_packet_info_s) next_by_hop;
    SLIST_ENTRY(trace_packet_info_s) next;
}
*trace_packet_info;		/* indexed by dport - dport */

/* list of packet containers */
static SLIST_HEAD(packets_s, trace_packet_info_s) trace_packets;
static int trace_packets_num = 0;

/* hop information, by ttl */
struct hop_info_s
{
    int num_sent;
    int all_sent, all_rcvd;
    struct timeval ts_last_sent;
    struct timeval ts_last_recv;	
    unsigned short state;
    unsigned short flags;
    SLIST_HEAD(hop_packets_s, trace_packet_info_s) packets;
}
*hop_info;

/* As the trace progresses, each hope will attempt
to work through the states one by one until it
receives an answer (2 attempts per state).
Whatever state "works" - will be then set up on 
following hops to continue from.
*/
#define HS_SEND_FIN         0x00
#define HS_SEND_SYN         0x01
#define HS_SEND_SYN_FIN		0x02    
#define HS_SEND_RST         0x04
#define HS_SEND_SYN_ACK     0x12
#define HS_SEND_ACK         0x16
#define HS_MAX              (HS_SEND_SYN)

#define HF_ENDPOINT         0x01

/* ToS (type of service) bits we can set on the IP datagram */

#define TOSMINDELAY         0x10
#define TOSMAXTHROUGH       0x08
#define TOSMAXRELIABLE      0x04
#define TOSMINCOST          0x02

#if defined(BSD_IP_STACK) && !defined(OPENBSD)
#define SCREWED_IP_LEN
#endif

#ifndef SCREWED_IP_LEN 
static u_int32_t
ip_cksum (const struct ip *ip)
{
    register const u_short *sp = (u_short *) ip;
    register u_int32_t sum = 0; 
    register int count;
    
    /*
     * No need for endian conversions.
     */
    for (count = ip->ip_hl * 2; --count >= 0;)
        sum += *sp++;
    while (sum > 0xffff)
        sum = (sum & 0xffff) + (sum >> 16);
    sum = ~sum & 0xffff;
    
    return (sum);
}
#endif

/* A standardized way of calculating checksums */
unsigned short 
udp_cksum (unsigned short *addr, signed int len) {
    
    unsigned short answer = 0;
    register unsigned short *w = addr;
    register int nleft = len;
    register int sum = 0; 
    
    while(nleft > 1) {
        sum += *w++;
        nleft -= 2;
    }
    
    if(nleft == 1) {
        *(unsigned char *)(&answer) = *(unsigned char *)w;
        sum += answer;
    }
    
    sum = (sum>>16) + (sum&0xffff);
    sum += (sum>>16);
    answer = ~sum;
    return (answer);
}

static u_int32_t
tcp_cksum (struct ip *ip, struct tcphdr *tcp, int payload_len)
{
    u_int32_t sum = 0; 
    register int count;
    u_short *sp;
    int old_ttl;
    
    sp = (u_short *) tcp;
    for (count = (sizeof (struct tcphdr) + payload_len) / 2; --count >= 0;)
        sum += *sp++;
    
    sp = (u_short *) & ip->ip_src;
    for (count = 2; --count >= 0;)
        sum += *sp++;
    
    sp = (u_short *) & ip->ip_dst;
    for (count = 2; --count >= 0;)
        sum += *sp++;
    
    old_ttl = ip->ip_ttl;
    ip->ip_ttl = 0;
    sp = (u_short *) & ip->ip_ttl;	/* protocol comes right after this */
    for (count = 1; --count >= 0;)
        sum += *sp++;
    ip->ip_ttl = old_ttl;
    
    sum += htons (sizeof (struct tcphdr));
    sum += payload_len;
    
    while (sum > 0xffff)
        sum = (sum & 0xffff) + (sum >> 16);
    sum = ~sum & 0xffff;
    
    return (sum);
}

static int
lft_resolve_port (const char *strport)
{
    
    struct servent *se;
    
    if ((se = getservbyname (strport, "tcp")))
        return (ntohs (se->s_port));
    
    return atoi (strport);
}

static char * 
strtolower (char *input) {
    char *iter = input;
    
    if (input == NULL)
        return NULL;
    
    while (*iter) {
        *iter = tolower (*iter);
        iter++;
    }
    return input;
}

void
do_auto_ports (char *hostname, int init_dport)
{
    /*
     *  Provides rudimentary auto-select of source and destination ports
     *  based on parsing the hostname supplied by the user
     *
     *  EXPECTS a string like "mail.example.com" and RETURNS the 
     *  auto-selected src and dst ports if changed or NULL if unchanged
     *
     *  Currently operates on user's input only as a call to
     *  gethostbyaddr() before tracing could slow us down too much
     */
    
    if (strlen(hostname) > 5) {
        
        char *mailservers[] = { "mail", "smtp", "mx", "relay", "inbound" };
        int cnt = 0;
        hostname = (char *) strtolower(hostname);
        
        for (cnt = 0 ; cnt < 5 ; cnt++)
        {
            if (strstr(hostname,mailservers[cnt]))
            {
                dport = 25;
                sport = 25;
                break;
            }
        }
        
        if (strncmp (hostname,"ftp",3) == 0) {
            dport = 21;
            sport = 19021;
        } else if (strncmp (hostname,"whois",5) == 0) {
            dport = 43;
            sport = 19043;
        } else if (strncmp (hostname,"imap",4) == 0) {
            dport = 143;
            sport = 19143;
        } else if (strncmp (hostname,"pop",3) == 0) {
            dport = 110;
            sport = 19110;
        } else if ((strncmp (hostname,"dns",3) == 0) || (strncmp (hostname,"ns",2) == 0) || (strncmp (hostname,"udns",4) == 0)) {
            dport = 53;
            sport = 53;
        } else if ((strncmp (hostname,"ntp",3) == 0) || (strncmp (hostname,"clock",5) == 0)) {
            dport = 123;
            sport = 123;
        }
        
        if (noisy && (dport != init_dport)) 
            printf ("Autoconfigured to source port %d, destination port %d.\n",sport, dport);
    }
}

void
print_host (struct in_addr addr)
{
    struct hostent *h;
    
    if (!resolve_names) {
        printf ("%s", inet_ntoa (addr));
    } else {
        h = gethostbyaddr ((void *) &addr, 4, AF_INET);
        if (h) {
            printf ("%s", h->h_name);
            if (!hostnames_only) 
                printf(" (%s)", inet_ntoa (addr));
        } else
            printf ("%s", inet_ntoa (addr));
    }
}

unsigned int
get_address(char *host)
{
    struct hostent *h;
    struct in_addr *addr;
    
    h = gethostbyname (host);
    if (!h) {
        fprintf (stderr, "LFT: Unknown host: %s\n", host);
        exit (1);
    }
    addr = (struct in_addr *) h->h_addr;
    return addr->s_addr;
}

#ifdef __CYGWIN__
/*
 The Windows/Cygwin version of this function uses the winsock built-in 
 WSAIoctl SIO_ROUTING_INTERFACE_QUERY to find the appropriate interface.
 */
char *
lft_getifforremote(const char *remote)
{
    struct hostent *h;
	struct sockaddr_in in, out;
	unsigned long nBytesReturned;
    
	/* Only do this once of course */
	if (sock < 0) {
		if ((sock = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
			perror ("socket"); 
			return NULL;
		}
	}
    
	/* Ask what i/f  the packet should go out on */
	in.sin_family = AF_INET;
	in.sin_port = 0;
	in.sin_addr.s_addr = get_address((char *)remote);
	if (SOCKET_ERROR != WSAIoctl(sock, SIO_ROUTING_INTERFACE_QUERY, &in, sizeof(in), &out, sizeof(out), &nBytesReturned, NULL, NULL))
		return lft_getifname(out.sin_addr);
    
	/* not found */
	return NULL;
}
#else
/* 
 The non-Windows version of this function uses connect() to 
 acquire a (UDP) socket to the target and uses the OS's decision 
 as to what interface is appropriate. 
 */
char *
lft_getifforremote(const char *remote)
{
    int sd;
    struct sockaddr_in sock;
    uint32_t socklen = sizeof(struct sockaddr_in);
    uint16_t p1;
    
    /* TODO: fill with a better RANDOM */
    
    p1 = rand() % 33525 + 32525;
    
    if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {perror("LFT: Socket trouble; unable to determine interface"); return 0;}
    sock.sin_family = AF_INET;
    sock.sin_addr.s_addr = get_address((char *) remote);
    sock.sin_port = htons(p1);
    if (connect(sd, (struct sockaddr *) &sock, sizeof(struct sockaddr_in)) == -1)
    {
        perror("LFT: UDP connect(); unable to determine interface");
        close(sd);
        return NULL;
    }
    
    memset(&sock, 0, sizeof(sock));
    if (getsockname(sd, (struct sockaddr*) &sock, &socklen) == -1) {
        perror("getsockname");
        close(sd);
        return NULL;
    }
    
    close(sd);
    
    return lft_getifname(sock.sin_addr);
}
#endif

static void
init_address (char *remote, char *pcap_dev)
{
    
    struct hostent *h;
    
    local_address.s_addr = lft_getifaddr (pcap_dev);	
    
    if (noisy) {
        print_host (local_address);
        if (random_source) {
            printf (":%d (pseudo-random)\n", sport);
        } else 
            printf (":%d\n", sport);
    }
    
    h = gethostbyname (remote);
    if (!h) {
        fprintf (stderr, "LFT: Unknown host: %s\n", remote);
        exit (1);
    }
    remote_address = *(struct in_addr *) h->h_addr;
}

static void
open_sockets ()
{
#ifdef __CYGWIN__
    int optval = 1;
    DWORD dwBytesRet = 2048;
#endif
#ifdef IP_HDRINCL
    int on = 1;
#endif
    int i;
#if defined(sun) || defined(__CYGWIN__)
    struct sockaddr_in local_bind;
#endif
    
#if defined(sun)
    send_sock = socket (PF_INET, SOCK_RAW, IPPROTO_IP);
#elif defined(BSD_IP_STACK)
    if (use_udp) 
        send_sock = socket (PF_INET, SOCK_RAW, IPPROTO_UDP);
    else 
        send_sock = socket (PF_INET, SOCK_RAW, IPPROTO_TCP);
#elif defined(__CYGWIN__)
    send_sock = socket (PF_INET, SOCK_RAW, IPPROTO_RAW);
    recv_sock = socket (PF_INET, SOCK_RAW, IPPROTO_IP);
    
    if (recv_sock < 0) {
        perror ("LFT: raw socket");
        exit (1);
    }
    local_bind.sin_addr = local_address;
    local_bind.sin_port = 0;
    local_bind.sin_family = AF_INET;
    if (bind(recv_sock, (struct sockaddr *)&local_bind, sizeof(local_bind)) < 0) {
        perror("winsock bind");
    }
    /* apparently the cygwin include files don't define this: */
# define SIO_RCVALL 0x98000001
    if (WSAIoctl(recv_sock, SIO_RCVALL, &optval, sizeof(optval), NULL, 0, &dwBytesRet, NULL, NULL) < 0)
        perror("LFT: WSAIoctl");
#else
    send_sock = socket (PF_INET, SOCK_RAW, IPPROTO_RAW);
#endif
    if (send_sock < 0) {
        perror ("LFT: raw socket");
        exit (1);
    }
#ifdef IP_HDRINCL
    if (setsockopt (send_sock, IPPROTO_IP, IP_HDRINCL, (char *) &on,
                    sizeof (on)) < 0) {
        perror ("LFT: IP_HDRINCL");
        exit (1);
    }
#endif
#ifdef sun
    local_bind.sin_addr = local_address;
    local_bind.sin_port = 0;
    local_bind.sin_family = AF_INET;
    if (bind (send_sock, (struct sockaddr *)&local_bind, sizeof (local_bind)) < 0) {
        perror ("LFT: bind");
        exit (1);
    }
#endif
    
    /* Prepare the trace packet (probe) */
    memset (&trace_packet, 0, sizeof (trace_packet));
    
    if (userlen == 0 && !use_udp) {
        /* If user doesn't supply length, default to zero-payload for TCP packets */
        trace_packet.payload_len = 0;
        trace_packet.payload = NULL;
    }
    else if (userlen == 0) {
        /* If payload length is not even (req'd by UDP), add padding */
        if ((sizeof(def_payload) % 2) != 0) {
            if (!(payload = malloc(sizeof (def_payload)+1)))
                perror("malloc()");
            bzero(payload,sizeof(payload));
            memcpy(payload,def_payload,sizeof(def_payload));
            trace_packet.payload_len = sizeof(def_payload) + 1;
            /* fill in a NULL as the padding of the last position */
            payload[trace_packet.payload_len - 1] = '\0';
        } else {
            if (!(payload = malloc(sizeof (def_payload))))
                perror("malloc()");    
            bzero(payload,sizeof(payload));
            memcpy(payload,def_payload,sizeof(def_payload));
            trace_packet.payload_len = sizeof(def_payload);
        } 
                
    } else {
        if (hostname_lsrr_size > 0) {
            if (use_udp) 
                userlen -= (sizeof (struct ip) + sizeof(struct udphdr) + trace_packet.lsrr.ipl_len + 1);
            else
                userlen -= (sizeof (struct ip) + sizeof(struct tcphdr) + trace_packet.lsrr.ipl_len + 1);
        } else {
            if (use_udp) 
                userlen -= (sizeof (struct ip) + sizeof(struct udphdr));
            else
                userlen -= (sizeof (struct ip) + sizeof(struct tcphdr));            
        }
        if (!(payload = malloc(userlen)))
            perror("malloc()");    
        bzero(payload,sizeof(payload));
        trace_packet.payload_len = userlen;      
    }
    
    payloadlen = trace_packet.payload_len;
    trace_packet.payload = payload;
    
    /* set up initial ip headers, etc. */
        
    if (hostname_lsrr_size > 0) {
        for (i = 0; i < hostname_lsrr_size; i++) {
            trace_packet.lsrr.data[i] = get_address(hostname_lsrr[i]);
        }
        trace_packet.lsrr.ipl_code = IPOPT_LSRR;
        trace_packet.lsrr.ipl_len = hostname_lsrr_size * 4 + 3;
        trace_packet.lsrr.ipl_ptr = 4;
    }
    trace_packet.ip_hdr.ip_v = 4;
    if (hostname_lsrr_size > 0) {
        trace_packet.ip_hdr.ip_hl = 6 + hostname_lsrr_size; /* 5 + 3byte lsrr + addresses + padding */
        if (use_udp) 
            trace_packet.ip_hdr.ip_len = sizeof (struct ip) + sizeof(struct udphdr) + trace_packet.lsrr.ipl_len + 1 + trace_packet.payload_len;
        else 
            trace_packet.ip_hdr.ip_len = sizeof (struct ip) + sizeof(struct tcphdr) + trace_packet.lsrr.ipl_len + 1 + trace_packet.payload_len;
    } else {
        trace_packet.ip_hdr.ip_hl = 5;
        if (use_udp) 
            trace_packet.ip_hdr.ip_len = sizeof (struct ip) + sizeof(struct udphdr) + trace_packet.payload_len;
        else
            trace_packet.ip_hdr.ip_len = sizeof (struct ip) + sizeof(struct tcphdr) + trace_packet.payload_len;
    }
    
#ifdef SCREWED_IP_LEN
    /*  trace_packet.ip_hdr.ip_len = sizeof (struct ip) + sizeof(struct tcphdr); */
#else
    trace_packet.ip_hdr.ip_len = htons (trace_packet.ip_hdr.ip_len);
#endif
#ifndef IPDEFTTL
#define IPDEFTTL 200
#endif
    trace_packet.ip_hdr.ip_ttl = IPDEFTTL;
    if (use_udp) 
        trace_packet.ip_hdr.ip_p = IPPROTO_UDP;
    else {
        trace_packet.ip_hdr.ip_p = IPPROTO_TCP;
        if (set_tos) 
            trace_packet.ip_hdr.ip_tos = TOSMINDELAY;
        trace_packet.tcp_hdr.th_win = htons (win_len);
        trace_packet.tcp_hdr.th_off = sizeof (struct tcphdr) / 4;
    }
    
    /* 
     * Init hop storage (array)
     * Init global packet info storage (SLIST)
     * Init per-hop packet info storage (SLIST)
     */
    hop_info_size = 255; /* can't be more than this */
    hop_info = malloc (sizeof (struct hop_info_s) * hop_info_size);
    for (i = 0; i < 255; i++) {
        bzero(&hop_info[i], sizeof(struct hop_info_s));
        hop_info[i].state = 0;
        SLIST_INIT(&(hop_info[i].packets));
    }
    
    SLIST_INIT(&trace_packets);
}

static unsigned int
new_seq()
{
    if (adaptive) {
        return rand();
    } else {
        return seq_start + trace_packets_num;
    }
}


static unsigned int
send_packet (short nhop, unsigned short ttl, unsigned int seq, unsigned char flags)
{
    struct sockaddr_in dest;
    unsigned int tseq=0;
    unsigned short tttl=0;
    struct sumh sum;
    char buf[maxpacklen+64];
    char* bptr = NULL;
    int blen = 0;
    char* s;
    
    /* we'll use gettimeofday() from checktimeouts 
     * instead of doing it again here               */
    /* gettimeofday (&now, NULL); */
    
    struct trace_packet_info_s	*pinfo = NULL;
    struct trace_packet_s *packet = NULL;
    
    if (!(pinfo = malloc(sizeof(struct trace_packet_info_s)))) {
        perror("malloc");
        exit(1);
    }
    
    bzero(pinfo, sizeof(struct trace_packet_info_s));
    packet = &(pinfo->packet);
    memcpy(packet, &trace_packet, sizeof(struct trace_packet_s));
    
    bptr = buf;
    
    dest.sin_family = AF_INET;
    dest.sin_addr = remote_address;
    dest.sin_port = 0;    
    
    if (seq == 0)
        tseq = new_seq();
    else
        tseq = seq;
    
    if (nhop == -1)
        tttl = ttl;
    else 
        tttl = nhop + 1;
    
    num_sent++;
    
    ts_last_sent = now;
    
    packet->ip_hdr.ip_ttl = tttl;
    packet->ip_hdr.ip_src = local_address;
    packet->ip_hdr.ip_dst = remote_address;
    
    if (noisy > 1 && !use_udp) {
        printf("SENT TCP  TTL=%d SEQ=%u FLAGS=%#x ( ", nhop+1, tseq, flags);
        if (flags & TH_RST)
            printf ("RST ");
        if (flags & TH_ACK)
            printf ("ACK ");
        if (flags & TH_SYN)
            printf ("SYN ");
        if (flags & TH_FIN)
            printf ("FIN ");
        printf(")\n");
    }
    else if (noisy > 1 && use_udp) {
        printf("SENT UDP  TTL=%d DPORT=%d\n", nhop+1, (dport + tttl));
    }
    else {
        /*  if (!nostatus)
        printf(">"); */
    }
    
    packet->ip_hdr.ip_sum = 0;
#if !defined(SCREWED_IP_LEN)
    packet->ip_hdr.ip_sum = ip_cksum (&(packet->ip_hdr));
#endif
        
    memcpy(bptr, &(packet->ip_hdr), sizeof(struct ip));
    bptr += sizeof(struct ip);
    if (packet->lsrr.ipl_len > 0) {
        memcpy(bptr, &(packet->lsrr), packet->lsrr.ipl_len + 1);
        bptr += (packet->lsrr.ipl_len + 1); /* PADDING !!! */
    }
    
    /* Layer-4 preparation */
    
    if (use_udp) {
        
        if (noisy > 5) 
            printf("Payload:  Length=%d  Contents=\"%s\"\n", packet->payload_len, packet->payload);
        /* Construct UDP (with payload) */        
        packet->udp_hdr.uh_dport = htons(dport + tttl);
        packet->udp_hdr.uh_sport = htons (sport);
        packet->udp_hdr.uh_ulen = htons ((sizeof (struct udphdr)) + packet->payload_len);
        /* truncated-udplength 0
            packet->udp_hdr.uh_ulen = 0;   
        packet->udp_hdr.th_flags = flags;
        */
                
        /* create pseudo-header for checksum calculation */
        sum.src=local_address.s_addr;
        sum.dst=remote_address.s_addr;
        sum.fill=0;
        sum.protocol=IPPROTO_UDP;
        sum.len=htons(sizeof(struct udphdr) + packet->payload_len);

        if(!(s=(char *)malloc(sizeof(struct sumh)+sizeof(struct udphdr)
                              +packet->payload_len)))
            perror("malloc()");
        memset(s,0,(sizeof(struct sumh)+sizeof(struct udphdr)
                    +packet->payload_len));
        memcpy(s,&sum,sizeof(struct sumh));
        memcpy(s+sizeof(struct sumh),&(packet->udp_hdr),sizeof(struct udphdr));
        memcpy(s+sizeof(struct sumh)+sizeof(struct udphdr),
               packet->payload,packet->payload_len);
        
        packet->udp_hdr.uh_sum=udp_cksum((unsigned short *)s, 
                                    sizeof(struct sumh)+sizeof(struct udphdr)+packet->payload_len);
        free(s);
        
        if (noisy > 5)
            printf("UDP Checksum = %#x\n",packet->udp_hdr.uh_sum);
        
#if defined(SOLARIS_LENGTH_IN_CHECKSUM)
        packet->udp_hdr.uh_sum = htons (sizeof (struct udphdr) + packet->payload_len);
#endif
        memcpy(bptr, &(packet->udp_hdr), sizeof(struct udphdr));
        bptr += sizeof(packet->udp_hdr);
        memcpy(bptr, packet->payload, packet->payload_len);
        blen = sizeof(struct ip) + packet->lsrr.ipl_len + sizeof(struct udphdr) + packet->payload_len;
    }
    
    else {
        /* Construct TCP (no payload needed) */
        if (noisy > 5) 
            printf("Payload:  Length=%d  Contents=\"%s\"\n", packet->payload_len, packet->payload);
        packet->tcp_hdr.th_dport = htons (dport);
        /*
         trace_packet.tcp_hdr.th_seq = htonl (seq_start + trace_packet_info_length);
         */
        packet->tcp_hdr.th_seq = htonl (tseq);
        packet->tcp_hdr.th_sport = htons (sport);
        packet->tcp_hdr.th_flags = flags;
        
#if defined(SOLARIS_LENGTH_IN_CHECKSUM)
        packet->tcp_hdr.th_sum = htons (sizeof (struct tcphdr)) + packet->payload_len;
#else
        packet->tcp_hdr.th_sum = 0;
        packet->tcp_hdr.th_sum = tcp_cksum (&(packet->ip_hdr), &(packet->tcp_hdr), packet->payload_len);
#endif
        if (noisy > 5)
            printf("TCP Checksum = %#x\n",packet->tcp_hdr.th_sum);
        
        memcpy(bptr, &(packet->tcp_hdr), sizeof(struct tcphdr));
        bptr += sizeof(packet->tcp_hdr);
        memcpy(bptr, &(packet->payload), packet->payload_len);
        blen = sizeof(struct ip) + packet->lsrr.ipl_len + sizeof(struct tcphdr) + packet->payload_len;
    }
    
    /* Packet is ready, fire away */
    if (sendto (send_sock, buf, blen, 0, (struct sockaddr *)&dest, sizeof (dest)) < 0) {
#ifdef __CYGWIN__
        perror ("sendto");
        if (!use_udp) {
            fprintf(stderr,"LFT:  Your platform may prevent you from using raw TCP sockets.\n");
            fprintf(stderr,"      Try UDP-based tracing instead using the \"-u\" option.\n");
        }
        exit (1);
#else
        perror ("sendto");
        fprintf(stderr,"LFT:  Your platform may prevent you from using raw TCP sockets.\n");
        fprintf(stderr,"      This could be the result of a local (host-based) firewall\n");
        fprintf(stderr,"      or a permissions problem on this binary or the BPF device.\n");
        if (adaptive)
            fprintf(stderr,"      You can try TCP-based tracing without using adaptive mode.\n");
        fprintf(stderr,"      You can try UDP-based tracing using the \"-u\" option.\n");
        exit (1);        
#endif
    }
    pinfo->hopno = nhop;
    if (use_udp) 
        pinfo->seq = dport + tttl;
    else 
        pinfo->seq = tseq;
    pinfo->sent = now;
    SLIST_INSERT_HEAD(&trace_packets, pinfo, next);
    trace_packets_num++;
    
    if (nhop != -1) {
        SLIST_INSERT_HEAD(&hop_info[nhop].packets, pinfo, next_by_hop);
        hop_info[nhop].num_sent++;
        hop_info[nhop].all_sent++;
        hop_info[nhop].ts_last_sent = now;
    }
    
    return tseq;
}

static double
timediff_ms (struct timeval prior, struct timeval latter)
{
    return
    (latter.tv_usec - prior.tv_usec) / 1000. +
    (latter.tv_sec - prior.tv_sec) * 1000.;
}

static unsigned int
send_hop (short nhop)
{
    struct hop_info_s *h = &hop_info[nhop];
    
    if (!adaptive)
        return send_packet (nhop , 0, 0, tcp_flags);
    
    if (h->state == HS_SEND_FIN) {
        return send_packet(nhop, 0, 0, TH_FIN);
    }
    
    if (h->state == HS_SEND_SYN) {
        return send_packet(nhop, 0, 0, TH_SYN);
    }
    
    if (h->state == HS_SEND_SYN_ACK) {
        return send_packet(nhop, 0, 0, HS_SEND_SYN_ACK);
    }    
    
    printf("Bad state %#x for hop %d\n", h->state, nhop);
    return -1;
}

static int
hop_state_up (short nhop)
{
    struct hop_info_s *h = &hop_info[nhop];
    
    if (h->state == HS_MAX)
        return -1;
    
    h->state++;
    /* This is a bad way to skip from SEND_SYN to SEND_SYN_ACK */
    if (h->state > HS_SEND_SYN && h->state < HS_MAX) 
        h->state = HS_SEND_SYN_ACK;
    
    h->num_sent = 0; /* for this state that is */
    return 0;
}

static int
hop_state_copy(short nhop)
{
    int i;
	
    if (noisy > 4)
        printf("Upping states of the hops following %d\n", nhop);
    for (i = (nhop+1); i <= 255; i++)
        if (hop_info[i].state < hop_info[nhop].state) {
            hop_info[i].state = hop_info[nhop].state;
            hop_info[i].num_sent = 0;
        }
            
            return 0;
}

void
finish ()
{
    int hopno;
    int maxhop;
    int reply, noreply;
    int noreply_hop_first = 0;
    int as_for_hop = 0;
    long elapsed = 0;
    struct trace_packet_info_s 	*tp;
    char *netname; 
    char ind;
    char *myApp = malloc((strlen(version) * sizeof(char)) + 1 + (strlen(appname) * sizeof(char)));
    struct ip_list_array ipaslist;
    memset(&tbuf, 0, sizeof(tbuf));
    gettimeofday (&trace_done_time, NULL);
    
    if (noisy > 3) printf ("Concluding with %d hops.\n", (num_hops+1));
    if (num_hops) {
        maxhop = num_hops;
        /* display all packets we received from this host */
        SLIST_FOREACH(tp, &trace_packets, next)
            if (tp->is_done)
                tp->hopno = maxhop;
    } else {
        maxhop = hop_info_length - 1;
    }
    
    if (num_hops && !nostatus && !noisy)
        printf ("T\n");
    else if (!noisy && !nostatus)
        printf ("\n");
    
    /* if user wants ASN resolution from pwhois/cymru/riswhois, do it in bulk */
    if (do_aslookup || do_netlookup) {
        if (!use_radb) {
            /* populate bulk ip_addr_list structure */
            for (hopno = ttl_min; hopno <= maxhop; hopno++) {
                SLIST_FOREACH(tp, &(hop_info[hopno].packets), next_by_hop)  {
                    if (tp->recv.tv_usec) {
                        ipaslist.ipaddr[as_for_hop] = tp->hopaddr;
                        as_for_hop++;
                        ipaslist.numItems = (as_for_hop);
                        break;
                    }
                }
            }
            if (use_cymru) {         /* use cymru bulk service */
                if (noisy > 1) printf ("Using Cymru for bulk AS resolution...\n");
                if (w_lookup_as_cymru_bulk(&ipaslist) != 0)
                    if (noisy) printf ("The Cymru lookup failed.\n");
            } else if (use_ris) {    /* use RIPE NCC RIS service */
                if (noisy > 1) printf ("Using RIPE NCC RIS for bulk AS resolution...\n");
                if (w_lookup_all_riswhois_bulk(&ipaslist) != 0)
                    if (noisy) printf ("The RIPE NCC RIS lookup failed.\n");
            } else {                 /* use pwhois bulk service */
                if (noisy > 1) printf ("Using Prefix WhoIs for bulk AS resolution...\n");
                if ((strlen(version) * sizeof(char)) + 1 + (strlen(appname) * sizeof(char)) < 254) {
                    strcat(myApp,appname); strcat(myApp," "); strcat(myApp,version);
                    strncpy(ipaslist.application,myApp,511);
                }
                if (w_lookup_all_pwhois_bulk(&ipaslist) != 0)
                    if (noisy) printf ("The Prefix WhoIs lookup failed.\n");
            }  
        } else {
            if (noisy > 1) printf ("Using RADB for in-line AS resolution...\n");
        }
    } 

free(myApp);

printf ("TTL LFT trace to ");

print_host (remote_address);
if (use_udp) {
    dport++;
    printf(":%d-%d/udp\n",dport, (dport + maxhop));
}
else 
  printf(":%d/tcp\n",dport);

noreply = 0;
reply = 0;
as_for_hop = 0;            /* this correlates the hopno to the asn stored in ipaslist */


for (hopno = ttl_min; hopno <= maxhop; hopno++) {
    struct in_addr last_hop;
    
    if (hop_info[hopno].all_rcvd != 0) {
        if (noreply == 1)
            printf("**  [neglected] no reply packets received from TTL %d\n", hopno);
        if (noreply > 1)
            printf("**  [neglected] no reply packets received from TTLs %d through %d\n", hopno - noreply + 1, hopno);
        
        noreply_hop_first = 0;
    }
    
    last_hop.s_addr = 0;
    if ((hop_info[hopno].state == HS_SEND_FIN) && (hop_info[hopno+1].state == HS_SEND_SYN) && (hop_info[hopno+1].ts_last_recv.tv_sec)) {
        printf("**  [firewall] the next gateway may statefully inspect packets\n");
    }

    if ((hop_info[hopno].state != HS_SEND_SYN_ACK) && (hop_info[hopno+1].state == HS_SEND_SYN_ACK) && (hopno == (num_hops - 1))) {
        printf("**  [firewall] the next gateway may implement a flag-based state filter\n");
    }    
    
    if ((hop_info[hopno].flags & HF_ENDPOINT) && (noreply >= ((maxhop - ttl_min)/2)) && num_hops > 3) {
        printf("**  [4.2-3 BSD bug?] the next gateway may errantly reply with reused TTLs\n");
    }
    
    if (hop_info[hopno].all_rcvd == 0) {
        reply = 0;
    } else {
        
        printf ("%2d ", hopno + 1);
        ind = ' ';
        
        SLIST_FOREACH(tp, &(hop_info[hopno].packets), next_by_hop) {
            
            if (tp->recv.tv_sec) {
                reply = 1;
                                                                
                if (last_hop.s_addr != tp->hopaddr.s_addr) {
                    ind = ' ';
                    if (do_aslookup) {
                        int asnumber = 0; 	/* init/clear the ASN */
                        if (use_radb) { 
                            /* using RADB/IRR */
                            asnumber = w_lookup_as(inet_ntoa(tp->hopaddr));
                            if (asnumber)
                                printf(" [%d]", asnumber);
                            else
                                printf(" [AS?]");
                        } else if (use_ris) {
                            /* using RIPE NCC RIS */
                            if ((asnumber = ipaslist.asn[as_for_hop]) > 0) 
                                /* if (asnumber) */
                                printf(" [%d]", asnumber);
                            else
                                printf(" [AS?]");
                        } else if (use_cymru) {
                            /* using cymru */
                            if ((asnumber = ipaslist.asn[as_for_hop]) > 0)
                                printf(" [%d]", asnumber);
                            else
                                printf(" [AS?]");
                        } else {
                            /* using pwhois by default */
                            if ((asnumber = ipaslist.asn[as_for_hop]) > 0 )
                                printf(" [%d]", asnumber);
                            else
                                printf(" [AS?]");
                        }
                    }
                                        
                    if (do_netlookup) {
                        if (!do_aslookup || (do_aslookup && !use_cymru && !use_radb)) {
                            if (strlen(netname = ipaslist.netName[as_for_hop]) > 1 )
                                printf(" [%s]", netname);
                            else
                                printf(" [Net?]");
                        } else {
                            netname = w_lookup_netname(inet_ntoa(tp->hopaddr));
                            if (netname)
                                printf(" [%s]", netname);
                            else
                                printf(" [Net?]");
                        }
                    }
                    
                    if (tp->icmp_type < -2 || tp->icmp_type > 17)
                        printf (" [icmp code %d]", tp->icmp_type);
                    else if (tp->icmp_type >= 0)
                        printf (" [%s]", icmp_messages[tp->icmp_type + 1]);                    
                    
                    if (tp->icmp_type == -1) {
                        printf(" [target");
                        if (target_open > 0) printf(" open");
                        printf("]");
                    }
                    printf(" ");
                     print_host (tp->hopaddr);
                    if (tp->icmp_type == -1) {
                        printf(":%d",dport);
                    }
                }
                
                last_hop = tp->hopaddr;
                printf ("%c%.1f", ind, timediff_ms(tp->sent, tp->recv));
            } else {
				/* only print out the information if this is a time differental record */
				if(tp->sent.tv_sec != 0 && tp->recv.tv_sec != 0) {
					printf ("%c*", ind);
				}
            }
            ind = '/';
        }
        printf ("ms\n");
    }
    if (reply) {
        noreply = 0;
        as_for_hop++;
    } else
        noreply++;
    
    reply = 0;
}

if (!num_hops) {
    if (target_anomaly)
        printf("**  [%d/tcp sequence anomaly from target]  Try advanced options (use -VV to see packets).\n", dport);
    else if (use_udp)
        printf("**  [%d-%d/udp no reply from target]  Use -VV to see packets.\n", dport, (dport + maxhop));
    else 
        printf("**  [%d/tcp no reply from target]  Try advanced options (use -VV to see packets).\n", dport);
}
if (timetrace) {
    gettimeofday (&now, NULL);
    (void)strftime(tbuf, sizeof(tbuf), time_format, (struct tm *)localtime((time_t *) &now.tv_sec));
    printf ("LFT trace finished at %s", tbuf);
    printf (" (%.2fs elapsed)",(timediff_ms(begin_time, now) / 1000));
    if (noisy) {
        printf ("\nTime spent tracing: %.2fs", 
                (timediff_ms(begin_time, trace_done_time) / 1000));
        if (resolve_names || do_aslookup || do_netlookup) 
            printf(", resolving: %.2fs", (timediff_ms(trace_done_time, now) / 1000));
    }
    printf("\n");
}
exit (0);
} 

void
check_timeouts ()
{
    int nhop;
    int need_reply = 0;
    int no_reply = 0;
    int last_return = 0;
    unsigned int tseq;
    
    tseq = 0;
    
    gettimeofday (&now, NULL);
    if (timediff_ms (ts_last_sent, now) < scatter_ms)
        return;			/* not ready to send another packet yet */
    
    for (nhop = ttl_min; nhop < hop_info_length; nhop++) {
        if (!hop_info[nhop].num_sent) {
            send_hop(nhop);
            return;
        }
    }
    
    for (nhop = ttl_min; nhop < hop_info_length; nhop++) {
        if (hop_info[nhop].num_sent <= retry_max && !hop_info[nhop].ts_last_recv.tv_sec) {
            if (noisy > 4) printf("No reply on TTL %d\n", nhop);
            if (timediff_ms (hop_info[nhop].ts_last_sent, now) >= timeout_ms) {
                /* we timed out waiting for this hop -- retry if we have any
                * more tries */
                if (hop_info[nhop].num_sent < retry_max) {
                    if (use_udp && !noisy && !nostatus) 
                        printf("*");
                    else if (!noisy && !nostatus) 
                        printf("*");
                    if (noisy > 2) printf("TTL %d timed out, (resending)\n",nhop);
                    send_hop(nhop);
                    return;
                } else {
                    if (!adaptive || hop_state_up(nhop)) {
                        if (noisy > 3) printf("TTL %d timed out, (giving up)\n",nhop);
                        no_reply++;
                    }
                }
            } else {
                need_reply++;		/* we have to wait for this one to timeout */
            }
        } else { /* have reply */
            last_return = nhop;
        }
    }
    
    if (noisy > 4) {
        printf ("| hilength %d, last_return %d\n", hop_info_length, last_return);
        printf ("| no_reply %d, ahead_limit %d\n", no_reply, ahead_limit);
        printf ("| num_hops %d, need_reply %d\n", num_hops, need_reply);
    }
    if (no_reply >= ahead_limit) {	/* we timed out. */
        if ((last_return + 3) * 2 < hop_info_length) {
            if ((need_reply < 3) && (num_rcvd < 2)) printf("LFT can\'t reliably round-trip.  Close-proximity filter in the way?\n");
            finish ();
        }
    }

if ((!num_hops || hop_info_length < num_hops || need_reply) && hop_info_length < ttl_limit) { 
    if (noisy > 4) printf("I still have unanswered hops.\n");
    if (need_reply >= ahead_limit) {
        if (noisy > 4) printf("I\'m too far ahead, returning.\n");
        return;			/* wait for some replies before we go on */
    }
    
    if (num_hops > 0 && hop_info_length >= num_hops) {
        if (noisy > 3) printf("I know the distance to the target, but I have gaps to fill.  Returning...\n");
        return;			/* we know how long the path is --
                         * wait to fill in the blanks      */
    }
    
    nhop = hop_info_length++;
    send_hop(nhop);
} else { 
    if (noisy >= 4) printf("Everyone either responded or timed out.  ");
    for (nhop = ttl_min; nhop < hop_info_length; nhop++) {
        if (hop_info[nhop].num_sent < retry_min && hop_info[nhop].num_sent <= retry_max /*&& hop_info[nhop].ts_last_recv.tv_sec*/) {
            send_hop(nhop);
            return;
        }
    }
    
    finish ();
} 
}

void
recv_packet (unsigned int seq, struct in_addr ipaddr, int icmp_type, const struct pcap_pkthdr *hdr)
{
    double ms;
    struct trace_packet_info_s *tp = NULL;
    
    /* Depending on the platform, we can use
     * the pcap header's timeval or we must call
       gettimeofday() for each packet  */
    
#ifdef __CYGWIN__
    gettimeofday (&now, NULL);
#else
    now.tv_sec = hdr->ts.tv_sec; 
    now.tv_usec = hdr->ts.tv_usec;
    /* gettimeofday (&now, NULL); */
#endif
    
    /* First, search every probe to find an exact sequence match */
    SLIST_FOREACH(tp, &trace_packets, next) {
        if (tp->seq == seq) {
            break;
        }
    } 

    /* Next, if no probes have an exact sequence match, look for an unincremented ACK */
    if (tp == NULL) {
        if (noisy > 3) printf("\nNo match in sequence check, looking for a match inside seq-.\n");
        SLIST_FOREACH(tp, &trace_packets, next) {       
            if (((tp->seq) == (seq +1)) && (icmp_type == -1))
                    break;
        }
    }
    
    /* Next, if no probes have an exact sequence match, look for an off-by-len */
    if (tp == NULL) {
        if (noisy > 3) printf("No match in seq-, looking for a match inside seq+len.\n");
        SLIST_FOREACH(tp, &trace_packets, next) {       
            if (((tp->seq) == (seq - payloadlen)) && (icmp_type == -1))
                break;
        }
    }    
    
    /* Last resort.  Catch any response from the target */
    if (tp == NULL) {
        if (noisy > 3) printf("No match in seq+len, looking for a match inside last resort loop.\n");
        SLIST_FOREACH(tp, &trace_packets, next) {       
            /* Special case: look for a response to our SYN_ACK */
            if (tp->packet.tcp_hdr.th_flags == HS_SEND_SYN_ACK) {
                if (!tp->recv.tv_sec) {
                    break;
                } 
            }            
            /* Truly the last resort: packet from the target with a wacky ACK sequence */
            if ((ipaddr.s_addr == remote_address.s_addr) && (tp->hopaddr.s_addr == 0) && (icmp_type == -1)) {
                target_anomaly = 1;
            }
        }
    }
    
    /* This packet is not even close, drop it and move on */
    if (!tp) {
        if (noisy)
            printf("(packet not meant for us, skip)\n");
        else
            if (!nostatus) printf("?");
        return;
    }
    
    if (tp->seq != seq) {
        if (((tp->seq) == (seq + 1)) && (icmp_type == -1)) {
            if (noisy > 1) {
                printf ("SRC=%s PTTL=%d\n", inet_ntoa (ipaddr), tp->hopno+1);
                printf ("Target\'s ACK was not incremented.  ");
            }
            /* return; */
        } else if (((tp->seq) == (seq - payloadlen)) && (icmp_type == -1)) {
            if (noisy > 1) {
                printf ("SRC=%s PTTL=%d\n", inet_ntoa (ipaddr), tp->hopno+1);
                printf ("Target\'s RST relates to the ISN + payload length (%d).  ", payloadlen);
            }
            /* return; */
        } else if ((ipaddr.s_addr == remote_address.s_addr) && (icmp_type == -1)) {
            if (noisy > 1) {
                printf ("SRC=%s PTTL=%d\n", inet_ntoa (ipaddr), tp->hopno+1);
                printf ("Target\'s ACK was way off.  ");
            }
            /* return; */
        }
    }
    
    if (tp->recv.tv_sec) {
        if (noisy)
            printf ("(duplicate packet, skip)\n");
        else
            if (!nostatus) printf("!");
        return;
    }
    
    if (noisy > 1) 
            printf ("SRC=%s PTTL=%d PSEQ=%u\n", inet_ntoa (ipaddr), tp->hopno+1, seq);
    else {
        if (!nostatus) printf(".");
    }
    
    /* increment received packet counter */
    num_rcvd++;
    
    tp->recv = now;
    if (tp->hopno != -1) {
        hop_info[tp->hopno].ts_last_recv = now;
        hop_info[tp->hopno].all_rcvd++;
        hop_state_copy(tp->hopno);
        /* indicate this hop has a sequence anomaly */
        
        if (icmp_type == -1)
            hop_info[tp->hopno].flags |= HF_ENDPOINT;
    }
    
    tp->hopaddr = ipaddr;
    tp->icmp_type = icmp_type;
    if (icmp_type != -2 && (!num_hops || num_hops > tp->hopno))
        if (break_on_icmp || (icmp_type == -1)) {
            if (tp->hopno != -1) { /* we set fake type -1 when we get actual
                * tcp packet in return - meaning destination */
                num_hops = tp->hopno;
                tp->is_done = 1;
                if (noisy > 1 && target_open == -1) printf ("Port %d/tcp appears to be closed; target sent RST.\n", dport);
                else if (noisy > 1 && target_open > 0) printf ("Port %d/tcp open; target attempted handshake.\n", dport);
            }
        }
            
            /* adjust scatter if we have fast reply times */
            ms = timediff_ms (tp->sent, tp->recv);
    scatter_ms = (scatter_ms * (ahead_limit - 1) + ms) / ahead_limit;
    
}

void process_packet (const u_char *packet, const struct pcap_pkthdr *hdr)
{
    const struct ip *ip, *orig_ip;
    const struct tcphdr *tcp;
    const struct udphdr *udp;
    const struct icmp *icmp;
    
    if (noisy > 4) printf("Received new data from packet capture; processing.\n");
    check_timeouts ();
    
    packet += skip_header_len;
    ip = (void *) packet;
            
    packet += 4 * ip->ip_hl;
            
    switch (ip->ip_p) {
        case IPPROTO_ICMP:
            orig_ip = ip;
            icmp = (void *) packet;
            if (icmp->icmp_type != ICMP_UNREACH && icmp->icmp_type != ICMP_TIMXCEED) {
                return;
            }
                ip = &icmp->icmp_ip;
            if (use_udp) {
                if (ip->ip_p != IPPROTO_UDP)
                    return;			/* not a response to our udp probe */
            } else {
                if (ip->ip_p != IPPROTO_TCP)
                    return;			/* not a response to our tcp probe */                
            }
            packet = (void *) ip;
            packet += 4 * ip->ip_hl;
            
            if (use_udp) {
                udp = (void *) packet;
                if (ntohs (udp->uh_sport) != sport || ip->ip_src.s_addr != local_address.s_addr || ip->ip_dst.s_addr != remote_address.s_addr) {
                    printf("Not for us\n");
                    return;			/* not for us */      
                }
                if (noisy > 2) {
                    printf("INCOMING IP:  SRC=%s PROTO=%d TTL=%d\n", inet_ntoa ((struct in_addr) orig_ip->ip_src), ip->ip_p, ip->ip_ttl);
                    printf("\\->ICMP+UDP:  ");
                    if (icmp->icmp_type == ICMP_TIMXCEED) 
                        printf("TTL exceeded; ");
                    else if (icmp->icmp_type == ICMP_UNREACH)
                        printf("unreachable; ");
                    printf("UDP SPORT=%d DPORT=%d\n", (ntohs (udp->uh_sport)), (ntohs (udp->uh_dport)));
                }
                if (noisy > 1)
                    printf ("RCVD ICMP ", ((ntohs (udp->uh_dport)) - dport));
                recv_packet (ntohs (udp->uh_dport) , orig_ip->ip_src,
                             (icmp->icmp_type == ICMP_TIMXCEED) ? -2 : icmp->icmp_code, hdr);                        
                
            } else {
                tcp = (void *) packet;
                if (ntohs (tcp->th_dport) != dport || ip->ip_src.s_addr != local_address.s_addr || ip->ip_dst.s_addr != remote_address.s_addr)
                    return;			/* not for us */                
                if (noisy > 2) {
                    printf("INCOMING IP:  SRC=%s PROTO=%d TTL=%d\n", inet_ntoa ((struct in_addr) orig_ip->ip_src), ip->ip_p, ip->ip_ttl);
                    printf("\\->ICMP+TCP:  ");
                    if (icmp->icmp_type == ICMP_TIMXCEED) 
                        printf("TTL exceeded; ");
                    else if (icmp->icmp_type == ICMP_UNREACH)
                        printf("unreachable; ");
                    printf("TCP SPORT=%d DPORT=%d\n", (ntohs (tcp->th_sport)), (ntohs (tcp->th_dport)));
                }
                if (noisy > 1) 
                    printf ("RCVD ICMP SEQ=%u ", ntohl (tcp->th_seq));
                recv_packet (ntohl (tcp->th_seq) , orig_ip->ip_src,
                             (icmp->icmp_type == ICMP_TIMXCEED) ? -2 : icmp->icmp_code, hdr);                        
            }
                return;
            
        case IPPROTO_TCP:
            /* check for RST reply */
            tcp = (void *) packet;
            if (!(tcp->th_flags & TH_RST) && !(tcp->th_flags & TH_ACK) && !(tcp->th_flags & TH_SYN)) 
                return;			/* not what we're looking for */
                
                if (ntohs (tcp->th_sport) != dport || ip->ip_src.s_addr != remote_address.s_addr || ip->ip_dst.s_addr != local_address.s_addr) {
                    return;			/* not the right connection */
                }
                    
                    if (noisy > 1) {
                        printf("RCVD TCP  FLAGS=%#x ( ",tcp->th_flags);
                        
                        if (tcp->th_flags & TH_SYN)
                            printf ("SYN ");
                        if (tcp->th_flags & TH_ACK)
                            printf ("ACK ");
                        if (tcp->th_flags & TH_FIN)
                            printf ("FIN ");
                        if (tcp->th_flags & TH_RST)
                            printf ("RST ");
                        
                        printf (") SEQ=%lu ACK=%lu ", ntohl (tcp->th_seq), ntohl (tcp->th_ack));
                    }
                    
                    /*if (ntohl(tcp->th_ack) < seq_start || ntohl(tcp->th_ack) > seq_start + trace_packet_info_length + 1)
                    return; * not for us */
                    
                    /* Check for SYN,ACK in response to determine if target is listening */
                    if ((tcp->th_flags & TH_ACK) && (tcp->th_flags & TH_SYN) && !(tcp->th_flags & TH_RST)) 
                        target_open++;
                    if ((tcp->th_flags & TH_ACK) && !(tcp->th_flags & TH_SYN) && (tcp->th_flags & TH_RST))
                        target_open = -1;
                        
                    recv_packet (ntohl (tcp->th_ack) - 1, ip->ip_src, -1, hdr);
            /*
             send_packet(-1, IPDEFTTL, ntohl(tcp->th_ack) + 1, TH_RST);
             */
            return;
            
        case IPPROTO_UDP:
            /* could be a probe we sent or something we don't support */
                return;			
            
        default:
            if (noisy > 3)
                printf("Incoming datagram contains unsupported protocol %d.\n", ip->ip_p);
    }
}

#ifndef __CYGWIN__
void
pcap_process_packet (u_char * user_data, const struct pcap_pkthdr *hdr,
                     const u_char * packet)
{
    process_packet(packet, hdr);
}
#endif

static void
usage (char *prog)
{
    fprintf (stderr,
             "\nLayer Four Traceroute (LFT)\n\n"
             "    - the alternative traceroute tool for network (reverse) engineers\n"
             "                                          visit http://www.pwhois.org\n\n"
             "Usage: %s [<options>] [<gateway> <...>] <target:port>\n"
             "\nMainstream Options:\n"
             "  -s <sport>       Source port number\n"
             "  -d <dport>       Destination port number (same as using target:port as target)\n"
             "  -z               Pseudo-randomize the source port number\n"
             "  -m <min>         Minimum number of probes to send per hop\n"
             "  -M <max>         Maximum number of probes to send per hop\n"
             "  -D <device|ip>   Use a specific device by name or IP address (\"en1\" or \"1.2.3.4\")\n"
             "  -L <length>      Set the length of the probe packet in bytes\n"
             "\nAdvanced Tracing Options:\n"
             "  -F               Use TCP FIN packets exclusively (defaults are SYN)\n"
             "  -e | -E          Use LFT's stateful engine to detect firewalls and path anomalies\n"
             "  -u               Use traditional UDP (probes) for tracing instead of TCP\n"
             "  -a <ahead>       Number of hops forward to query before pausing to wait for replies\n"
             "  -c <scatter ms>  Minimum number of milliseconds between subsequent probes\n"
             "  -t <timeout ms>  Maximum RTT to wait before assuming packet was dropped\n"
             "  -l <min ttl>     Minimum TTL to use on outgoing packets (skips close-proximity hops)\n"
             "  -H <ttl>         Maximum number of hops to traverse (max TTL of packets)\n"
             "  -q <sequence>    Set the initial sequence number (ISN) manually\n"
             "  -I               Set the ToS field in the IP packet to minimize-delay\n"
             "  -i               Disable \"stop on ICMP\" other than TTL expired\n"
             "\nResolution Options:\n"
             "  -n               Display hosts numerically; disable use of the DNS resolver\n"
             "  -h               Display hosts symbolically; suppress IP address display\n"
             "  -N               Display network or AS names where appropriate\n"
             "  -A               Display AS numbers resolved by Prefix WhoIs\n"
             "\nAdvanced Resolution Options:\n"
             "  -r               Use RIPE NCC's RIS to resolve ASNs instead of Prefix WhoIs\n"
             "  -R               Use the RADB to resolve ASNs instead of Prefix WhoIs\n"
             "  -C               Use Cymru to resolve ASNs instead of Prefix WhoIs\n"
             "\nVerbosity Options and Status:\n"
             "  -T               Use execution timers and summarize where LFT spent its time\n"
             "  -U               Display all times in UTC (GMT0).  Activates -T option automatically\n"
             "  -S               Disable the status bar (only show the completed trace)\n"
             "  -V               Display verbose/debug output.  Use more \'V\'s for additional detail\n"
             "  -v               Display LFT's version information and exit\n"  
             "\n"
             "Default is: %s -s %d -d %d -m %d -M %d -a %d -c %.0f -t %d -H %d \n\n",
             prog, prog, sport, dport, retry_min, retry_max, ahead_limit,
             scatter_ms, timeout_ms, ttl_limit);
    exit (1);
}

static void
show_version (void)
{
    fprintf (stderr, "\n"
             "Layer Four Traceroute (LFT) - version %s %s\n\n    - the alternative traceroute tool for network (reverse) engineers\n\n"
             "    Compile-time options:\n\n"
#if defined(DARWIN)
             "      + Darwin (or MacOS)\n"
#endif
#if defined(OPENBSD)
             "      + OpenBSD\n"
#endif
#if defined(BSD_IP_STACK)
             "      + BSD IP stack\n"
#endif
#if defined(BSD)
             "      + BSD platform\n"
#endif
#if defined(linux)
             "      + Linux platform\n"
#endif
#if defined(sun)
             "      + SUN platform\n"
#endif
#if !defined(sun) && !defined(linux) && !defined(BSD_IP_STACK) && !defined(OPENBSD)
             "      (unknown architecture)\n"
#endif
#if defined(SCREWED_IP_LEN)
             "      + IP length big-endian\n"
#else
             "      + IP length little-endian\n"
#endif
#if defined(IP_HDRINCL)
             "      + Full IP headers for raw sockets\n"
#else
             "      + Without IP header inclusion\n"
#endif
             "      + " HOST_SYSTEM_TYPE "\n"
             "\n", version, version_date);
    exit (0);
}

#ifdef __CYGWIN__
void cygwin_process()
{
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(recv_sock, &fds);
    struct timeval tm;
    tm.tv_sec = 0;
    tm.tv_usec = 100000;
    if (select(recv_sock+1, &fds, 0, 0, &tm) < 0) {
        perror("select");
        exit(1);
    }
    if (FD_ISSET(recv_sock, &fds)) {
        /* read packet */
        char packetbuf[2048];
        memset(packetbuf, 0, sizeof(packetbuf));
        int nread = recv(recv_sock, packetbuf, sizeof(packetbuf), 0);
        if (nread <= 0) {
            perror("read");
            exit(1);
        }
        process_packet(packetbuf, NULL);
    }
}

#endif

extern int
main (int argc, char **argv)
{
#ifndef __CYGWIN__
    char ebuf[PCAP_ERRBUF_SIZE];
    static pcap_t *pd;
#endif
    char *pcap_dev, *userdev = NULL;
    int ch;
    int use_fins = 0;
    char *cp;
    struct timeval tb;
    unsigned int bpfimmflag;
    
    setbuf (stdout, NULL);
    
    while ((ch = getopt (argc, argv, "Aa:Cc:D:d:EeFH:hIiL:l:M:m:Nnp:q:RrSs:Tt:UuVvw:z")) != EOF)
        switch (ch) {
            case 'F':
                if (adaptive) {
                    fprintf (stderr,
                             "LFT warning: TCP flags are selected automatically using (-E) option.\n             Ignoring FINs-only (-F) option and using adaptive Engine.\n");
                    break;
                } 
                tcp_flags = TH_FIN;
                use_fins = 1;
                dport = 25000;
                auto_ports = 0;
                break;
            case 'h':
                if (!resolve_names) {
                    fprintf (stderr,
                             "LFT warning: I can't display hostnames (-h) unless I resolve them.\n             Ignoring your request to display hostnames exclusively.\n");
                } else hostnames_only = 1;
                break;
            case 'u':
                use_udp = 1;
                if (adaptive) {
                    fprintf (stderr,
                             "LFT warning: Disabling adaptive mode while using UDP.\n");
                    adaptive = 0;
                }        
                if (use_fins) {
                    fprintf (stderr,
                            "LFT warning: Disabling FINs-only mode while using UDP.\n");
                    use_fins = 0;       
                }
                if (!dflag) 
                    dport = (start_dport - 1);  
                auto_ports = 0;
                break;
            case 'r':
                if (use_cymru) {
                    fprintf (stderr,
                             "LFT warning: Only one ASN lookup source may be used--you selected Cymru.\n             Ignoring your request to use RIPE NCC RIS.\n");
                    break;
                } else if (use_radb) {
                    fprintf (stderr,
                             "LFT warning: Only one ASN lookup source may be used--you selected RADB.\n             Ignoring your request to use RIPE NCC RIS.\n");
                    break;
                }
                use_radb = 0;
                use_ris = 1;
                use_cymru = 0;
                do_aslookup = 1;
                break;
            case 'R':
                if (use_cymru) {
                    fprintf (stderr,
                             "LFT warning: Only one ASN lookup source may be used--you selected Cymru.\n             Ignoring your request to use RADB.\n");
                    break;
                } else if (use_ris) {
                    fprintf (stderr,
                             "LFT warning: Only one ASN lookup source may be used--you selected RIPE NCC RIS.\n             Ignoring your request to use RADB.\n");
                    break;
                }
                
                use_radb = 1;
                use_cymru = 0;
                use_ris = 0;
                do_aslookup = 1;
                break;
            case 'C':
                if (use_radb) {
                    fprintf (stderr,
                             "LFT warning: Only one ASN lookup source may be used--you selected RADB.\n             Ignoring your request to use Cymru.\n");
                    break;
                } else if (use_ris) {
                    fprintf (stderr,
                             "LFT warning: Only one ASN lookup source may be used--you selected RIPE NCC RIS.\n             Ignoring your request to use Cymru.\n");
                    break;
                }
                use_cymru = 1;
                use_radb = 0;
                use_ris = 0;
                do_aslookup = 1;
                break;
            case 'd':
                ++dflag;
                if (use_udp && atoi (optarg) > (65535 - ttl_limit)) {
                    fprintf (stderr,
                             "LFT warning: Starting UDP port %d is too high.  Will start with %d instead.\n", atoi(optarg), (65535 - ttl_limit));
                    dport = (65535 - ttl_limit) - 1;
                } else 
                    dport = (atoi (optarg)) - 1;
                auto_ports = 0;
                break;
            case 'L':
                if (atoi (optarg) > maxpacklen) {
                    fprintf (stderr,
                             "LFT warning: Packet length %d is too high.  Will use %d instead.\n", atoi(optarg), maxpacklen);
                    userlen = maxpacklen;
                } else if (atoi (optarg) < minpacklen) {
                    fprintf (stderr,
                             "LFT warning: Packet length %d is too low.  Using default instead.\n", atoi(optarg));
                    userlen = 0;
                } else 
                    userlen = (atoi (optarg));
                break;                
            case 'q':
                seq_start = atol (optarg);
                break;
            case 'w':
                win_len = atoi(optarg);
                break;
            case 'm':
                retry_min = atoi (optarg);
                break;
            case 'M':
                retry_max = atoi (optarg);
                break;
            case 'N':
                do_netlookup = 1;
                break;
            case 'A':
                do_aslookup = 1;
                break;
            case 'n':
                if (!hostnames_only) {
                    resolve_names = 0; 
                } else {
                    fprintf (stderr,
                             "LFT warning: I can't display hostnames (-h) unless I resolve them.\n             Ignoring your request not to resolve DNS.\n");
                }
                break;
            case 'T':
                timetrace = 1;
                break;
            case 's':
                if (random_source) {
                    fprintf (stderr,
                             "LFT warning: You already asked to use a random source port.\n             Ignoring request to set specific source port.\n");
                    break; 
                }
                sport = lft_resolve_port (optarg);
                break;
            case 'E':
                if (use_udp) {
                    fprintf (stderr,
                             "LFT warning: Disabling adaptive mode while using UDP.\n");
                    break;
                }                        
                if (use_fins) {
                    fprintf (stderr,
                             "LFT warning: TCP flags are selected automatically using (-E) option.\n             Ignoring adaptive Engine (-E) option and using FINs-only (-F).\n");
                    break;
                }
                adaptive = 1;
                break;
            case 'e':
                if (use_fins) {
                    fprintf (stderr,
                             "LFT warning: TCP flags are selected automatically using (-e) option.\n             Ignoring adaptive Engine (-e) option and using FINs-only (-F).\n");
                    break;
                }
                if (use_udp) {
                    fprintf (stderr,
                             "LFT warning: Disabling adaptive mode while using UDP.\n");
                    break;
                }                        
                adaptive = 1;
                break;
            case 'S':
                nostatus = 1;
                break;
            case 'D':
            {
                if (strlen (optarg) > max_net_dev_input) {
                    fprintf (stderr,
                             "Net interface names are limited to %d characters. (e.g., \"eth0\" or \"ppp0\" or \"10.10.10.10\")\n",
                             max_net_dev_input);
                    exit (1);
                }
                userdevsel = 1;
                userdev = optarg;
            }
                break;
            case 'a':
                ahead_limit = atoi (optarg);
                break;
            case 'c':
                scatter_ms = atoi (optarg);
                if (scatter_ms < 1)
                    scatter_ms = 1;
                    if (scatter_ms > 100)
                        scatter_ms = 100;
                        break;
            case 't':
                timeout_ms = atoi (optarg);
                break;
            case 'H':
                if (atoi (optarg) > 255) 
                    ttl_limit = 255; else
                        ttl_limit = atoi (optarg);
                break;
            case 'l':
                ttl_min = atoi(optarg);
                hop_info_length = ttl_min;
                if (ttl_min > 0)
                    ttl_min--;
                    break;
            case 'i':
                break_on_icmp = 0;
                break;
            case 'I':
                set_tos = 1;
                break;
            case 'v':
                show_version ();
                break;
            case 'U':   /* show all times in UTC */
#if defined(sun)
                if (putenv("TZ=GMT0") == -1) {
                    fprintf(stderr, "%s: Unable to set TZ to UTC.",appname);
                }                
#else
                if (setenv("TZ", "GMT0", 1) == -1) {
                    fprintf(stderr, "%s: Unable to set TZ to UTC.",appname);
                }
#endif
                timetrace = 1;
                break;                
            case 'V':
                noisy++;
                nostatus = 1; 
                break;
            case 'z':
                random_source = 1;
                /* Yes, this is a ridiculous randomizer, but it's adequate */
                sport = rand()%32525+32525;
                break;
            default:
                usage (argv[0]);
        }
            
            if ((argc - optind) < 1)
                usage (argv[0]);
    
    
    if (noisy)
        printf ("Layer Four Traceroute (LFT) version %s", version);
    if (noisy > 1) 
        printf (" ... (verbosity level %d)",noisy);
    if (noisy)    
        printf ("\n");
    
    gettimeofday (&tb, NULL);
    /* eventually this might want to use /dev/urandom or
        * something on machines that have it.  otherwise,
        * this does a fairly decent job of using the system
        * clock.
        *
        * multiply tv_usec (range 0-1000000) to be in range 0-2^31,
        * and xor to randomize the high bits of tv_sec that don't
        * change very much.
        */
    srand(tb.tv_sec ^ (tb.tv_usec * 2147));
    
    hostname = argv[optind++];
    hostname_lsrr_size = 0;
    while (optind < argc) {
        hostname_lsrr[hostname_lsrr_size++] = argv[optind++];
        if (hostname_lsrr_size > IP_LSRR_DEST) {
            fprintf(stderr, "LFT: Too many LSRR hosts - maximum is 8\n");
            exit(1);
        }
    }
    if (hostname_lsrr_size > 0) {
        hostname_lsrr[hostname_lsrr_size++] = hostname;
        hostname = hostname_lsrr[0];
    }
    
    /* allow hostname:port if -d not specified and not using UDP */
    if ((cp = strchr (hostname, ':'))) {
        if (!dflag) {
            *cp++ = '\0';
            dport = lft_resolve_port (cp);
            if (use_udp) {
                dport = (lft_resolve_port (cp)) - 1;
                if (dport > (65535 - ttl_limit)) {
                    dport = (65535 - ttl_limit) - 1;
                    fprintf (stderr,
                             "LFT warning: Starting UDP port %d is too high.  Will start with %d instead.\n", dport, (65535 - ttl_limit));
                }
            }
            auto_ports = 0;
        }
    }
    
    if (auto_ports) 
        do_auto_ports(hostname, dport);
    
    if (do_netlookup || do_aslookup)
        w_init();                                                   /* initialize the whois framework */
    
#ifdef __CYGWIN__
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 1), &wsaData) != 0) {
        perror("WSAStartup");
        exit(1);
    }
#endif
    
    /* if not given network interface, select one automatically */
    if (!userdevsel) {
        pcap_dev = lft_getifforremote(hostname);
#ifndef __CYGWIN__
        if (pcap_dev == NULL) 
            pcap_dev = pcap_lookupdev (ebuf);
        if (pcap_dev == NULL) {
            fprintf (stderr, "LFT: %s\n", ebuf);
            exit (1);
        }
#else 
        if (pcap_dev == NULL) {
            fprintf (stderr, "LFT: Failed to discover an appropriate interface.\n");
            exit (1);
        }
#endif
    } else {
        struct in_addr addr;
        if (inet_aton(userdev, &addr)) {
            /* specified by ip address -- look up device. */
            pcap_dev = lft_getifname(addr);
            if (!pcap_dev) {
                fprintf (stderr, "LFT: Unable to locate a local interface with IP address %s\n", userdev);
                exit(1);
            }
        } else
            pcap_dev = userdev;
    };
    
#ifdef __CYGWIN__
    skip_header_len = 0;
#else
    if (!strncmp (pcap_dev, "lo:", 3) || !strcmp (pcap_dev, "lo")) {
        pcap_dev = "ppp0";
        skip_header_len = 0;
    } else if (!strncmp (pcap_dev, "ppp", 3))
        skip_header_len = 0;
    else				/* assume ethernet */
        skip_header_len = sizeof (struct ether_header);
#endif
    if (noisy)
        printf ("Using device %s, ", pcap_dev);
#ifndef __CYGWIN__
    pd = pcap_open_live (pcap_dev, 1600, 0, 20, ebuf);
    if (!pd) {
        fprintf (stderr, "The network device \"%s\" isn\'t available to LFT.  Try another or fix:\nERROR: %s\n", pcap_dev, ebuf);
        exit (1);
    }
#ifdef BSD_IP_STACK
#ifndef DARWIN
    bpfimmflag = 1;
    if ( ioctl(pd->fd, BIOCIMMEDIATE, &bpfimmflag) < 0) {
        fprintf(stderr, "BIOCIMMEDIATE: %s",
                pcap_strerror(errno));
    }
#endif
#endif
#endif
    
    init_address (hostname, pcap_dev);
    
    if (!seq_start && !use_udp) {
        seq_start = rand();
        if (noisy > 3) printf("Initial Sequence Number = %d\n", seq_start);
    }
    
    open_sockets ();
    
#ifndef __CYGWIN__
    setuid (getuid ());
#endif
    
    if (adaptive) {
        if (retry_min < 2)
            retry_min = 2;
    }
    
    if (use_udp) {
        if (retry_min > 2)
            retry_min = 2;
    }

    if (retry_max < retry_min)
        retry_max = retry_min;
    
    gettimeofday (&begin_time, NULL);
    if (timetrace) {
        memset(&tbuf, 0, sizeof(tbuf));
        (void)strftime(tbuf, sizeof(tbuf), time_format, (struct tm *) localtime((time_t *) &begin_time.tv_sec));
        printf ("LFT trace started at %s\n", tbuf);
    }
    
    if (!nostatus && !noisy)
        printf("Tracing ");
    
#ifdef __CYGWIN__
    for (;;) {
        cygwin_process();
        check_timeouts();
    }
#else
    while (pcap_dispatch (pd, -1, pcap_process_packet, 0) >= 0) {
        if (noisy > 6) printf("Left dispatch.\n");
        check_timeouts ();
    }
#endif
    
    return 0;
}

